// Implementation of the LMTEdge class.

#include <stdlib.h>
#include <float.h>
#include <vector>
 
#include "LMTPoint.h"
#include "LMTEdge.h"
#include "util.h"

// Initialize the group counter to 0.
int LMTEdge::groupCounter = 0;

// Get the next edge in the cyclic ordering of edges around an endpoint,
// performing lazy deletion.
LMTEdge* LMTEdge::getNext() {
#ifdef DEBUG
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  log4cpp::NDC::push("getNext");
  logger.infoStream() << " Next for " << *this << ".";
#endif
  // Follow the next-pointers to the first possible edge and remove
  // impossible edges from the list.
  LMTEdge* result = this->next;
  while (result->status == impossible) {
#ifdef DEBUG
  logger.infoStream() << " Skipping " << *result << ".";
#endif
    result = result->next;
    this->next = result;
  }
#ifdef DEBUG
  logger.infoStream() << " Result " << *result << ".";
  log4cpp::NDC::pop();
#endif
  return result;
}

// Reset the triangle to the left of the half-edge.
void LMTEdge::reset() {
  this->trianglesLeft = true;
  this->i = this->getNext();
  this->j = this->dstart;
}

// Advance the triangle to the left of the half-edge.
bool LMTEdge::advance() {
#ifdef DEBUG
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  log4cpp::NDC::push("advance");
  logger.infoStream() << " Advancing " << *this << ".";
#endif
  // Repeat until the right edge meets this edge or until the next triangle
  // has been found. 
  do {
    // If the right edge has been moved already,
    // move the left edge until it is to the left of the right edge or until
    // it is not to the left of this edge any more.
    while (this->j != this->dstart && this->i->dest->isLeftOf(this) &&
	   !this->i->dest->isLeftOf(this->j)) {
      this->i = this->i->getNext();
      if (this->i->dest == this->j->dest) {
	break;
      }
    }
#ifdef DEBUG
  logger.infoStream() << " I advanced to " << *(this->i->dest) << ".";
#endif
    // If there are no more left edges left or if we already have found
    // a triangle, abort.
    if (!this->i->dest->isLeftOf(this) || this->i->dest == this->j->dest) {
#ifdef DEBUG
      logger.infoStream() << " Abort!";
#endif
      break;
    }
    // If the right edge has not been moved yet, move it for the first time.
    if (this->j == this->dstart) {
      this->j = this->j->getNext();
    } else {
      // Otherwise, move the right edge until it meets the left edge or goes
      // past it.
      while (this->j->dest != this->rev->dest && 
	     this->i->dest->isLeftOf(this->j)) {
	this->j = this->j->getNext();
      } 
    }
#ifdef DEBUG
  logger.infoStream() << " J advanced to " << *(this->j->dest) << ".";
#endif
  } while (this->j->dest != this->rev->dest && this->j->dest != this->i->dest);
  // If the left edge went to the right of this edge or the right
  // edge met this edge, we went too far. In this case, there are no 
  // more triangles left.
  if (this->j->dest == this->rev->dest || !this->i->dest->isLeftOf(this)) {
    this->trianglesLeft = false;
  } else {
    this->trianglesLeft = true;
  }
#ifdef DEBUG
  logger.infoStream() << "Triangle advanced to " << *(this->j->dest) << ":"
		      << *(this->i->dest)
		      <<" (" << this->trianglesLeft << ").";
  log4cpp::NDC::pop();
#endif
  return this->trianglesLeft;
}

// Check the certificate for this edge
bool LMTEdge::checkCertificate() {
#ifdef DEBUG
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  log4cpp::NDC::push("checkCertificate");
  logger.infoStream() << "Checking certificate for " << *this << ".";
#endif
  // Initialize the triangles, if necessary.
  if (!this->certificateChecked) {
    this->reset();
    this->rev->reset();
#ifdef DEBUG
    logger.infoStream() << "Initialized to " << *(this->i) << " and "
		       << *(this->j) << ".";
    logger.infoStream() << "Reverse set to " << *(this->rev->i) << " and "
		       << *(this->rev->j) << ".";
#endif
    this->rev->advance();
    this->certificateChecked = true;
  }
  bool found = false;
  // Advance the triangles, until a certificate is found or no more
  // triangles remain.
  do {
    this->advance();
    if (!this->rev->trianglesLeft) {
#ifdef DEBUG
      logger.infoStream() << "No certificate found.";
#endif
      break;
    }
    if (!this->trianglesLeft) {
      this->rev->advance(); this->reset();
    } else {
#ifdef DEBUG
      logger.infoStream() << "Checking, whether " << *(this->rev->i->dest)
			  << " : " << *(this->i->dest) << " is a certificate.";
#endif
      // A quadrangle is a certificate iff it is not convex or this edge is
      // the shorter diagonal.
      if (!(isLeftOf(*(this->rev->i->dest), *(this->i->dest), 
		    *(this->rev->dest)) 
	   && isRightOf(*(this->rev->i->dest), *(this->i->dest), 
			*(this->dest))) ||
	  this->length <= lengthOf(*(this->rev->i->dest), *(this->i->dest))) {
#ifdef DEBUG
	logger.infoStream() << "Found Certificate.";
#endif
	found = true;
	break;
      }
    }
  }  while(true);
#ifdef DEBUG
  log4cpp::NDC::pop();
#endif
  return found;
}

// Check whether this edge intersects with edge.
bool LMTEdge::intersects(LMTEdge *edge) {
  return ((this->dest->isRightOf(edge) && this->rev->dest->isLeftOf(edge) &&
       edge->dest->isLeftOf(this) && edge->rev->dest->isRightOf(this)) ||
      (this->dest->isLeftOf(edge) && this->rev->dest->isRightOf(edge) &&
       edge->dest->isRightOf(this) && edge->rev->dest->isLeftOf(this)));
}

// Put all the edges whose certificates become invalid when 
// this edge becomes impossible back on the stack.
void LMTEdge::restackEdges(std::vector<LMTEdge *> &edgeStack) { 
#ifdef DEBUG
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  log4cpp::NDC::push("restackEdges");
#endif
  // Check all the triangles to the left of this edge.
  this->reset();
  while (this->advance()) {
    // If this edge is part of the certificate of one of the two triangle
    // edges, put the triangle edge back onto the stack, if necessary.
    if (this->i->rev->j == this && !this->i->onStack) {
      edgeStack.push_back(this->i);
      this->i->onStack = this->i->rev->onStack = true;
#ifdef DEBUG
      logger.infoStream() << "Restacking " << *(this->i) << ".";
#endif
    }
    if (this->j->i == this->rev && !this->j->onStack) {
      edgeStack.push_back(this->j);
      this->j->onStack = this->j->rev->onStack = true;
#ifdef DEBUG
      logger.infoStream() << "Restacking " << *(this->j) << ".";
#endif
    }
  }
#ifdef DEBUG
  log4cpp::NDC::pop();
#endif
}

// Check whether there are edges crossing this edge.
bool LMTEdge::checkCrossingEdges() {
#ifdef DEBUG
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  log4cpp::NDC::push("checkCrossingEdges");
  logger.infoStream() << "Checking crossing edges for " << *this << ".";
#endif
  bool found = false;
  // Check all the edges to the left of this edge.
  LMTEdge *u = this->getNext();
  while (u->dest->isLeftOf(this) && !found) {
    LMTEdge *v = u->rev;
    // For all edges to going out of the endpoint of u, check whether they
    // intersect this edge. 
    v = v->getNext();
    while (this->dest->isLeftOf(v) && v->dest != this->rev->dest) {
      if (this->intersects(v)) {
#ifdef DEBUG
	logger.infoStream() << "  Intersected by " << *v << ".";
#endif
	found = true;
	break;
      }
      v = v->getNext();
    }
    u = u->getNext();
  }
#ifdef DEBUG
  logger.infoStream()  
    <<(found ? " Intersection found. " : "No intersection found");
  log4cpp::NDC::pop();
#endif
  return found;
}

// Scan this edge during the dynamic programming step.
void LMTEdge::scan(int /*n*/) {
#ifdef DEBUG
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  log4cpp::NDC::push("scan");
  logger.infoStream() << "Scanning " << *(this) << ".";
#endif
  // Check all the triangles to the left of this edge.
  this->reset();
  this->polyWeight = DBL_MAX;
  while (this->advance()) {
#ifdef DEBUG
    int distI = min((this->i->destIndex - this->rev->destIndex + n) % n,
		    (this->rev->destIndex - this->i->destIndex + n) % n);
    int distJ = min((this->i->destIndex - this->destIndex + n) % n,
		    (this->destIndex - this->i->destIndex + n) % n);
    logger.infoStream() << "Dist: " << distI << " " << distJ << " " <<(this->j->rev->polyWeight);
#endif
    if (this->i->polyWeight != DBL_MAX && this->j->rev->polyWeight != DBL_MAX) {
      // Compute the weight of the next position.
      long double nextWeight = this->i->polyWeight + this->j->rev->polyWeight;
#ifdef DEBUG
      logger.infoStream() << "nextWeight " << nextWeight << ".";
#endif
      // Add the lenghts of the triangle edges, if they do not lie on
      // the edge of the polygon.
      if (this->i->status != certain) {
	nextWeight += this->i->length;
      }
      if (this->j->status != certain) {
	nextWeight +=  this->j->length;
      }
#ifdef DEBUG
      logger.infoStream() << "nextWeight " << nextWeight << ".";
#endif
      // Keep track of the current minimum weight.
      if (nextWeight < this->polyWeight) {
	this->polyWeight = nextWeight;
	leftPoly = this->i;
	rightPoly = this->j->rev;
      }
    }
  }
#ifdef DEBUG
  log4cpp::NDC::pop();
#endif
}

// Initialize this edge for traversal during dynamic programming.
void LMTEdge::traverse(const int counter) {
  this->rev->destIndex = counter;
  this->polyWeight = this->rev->polyWeight = 0;
  this->leftPoly = this->rightPoly = NULL;
  this->rev->leftPoly = this->rev->rightPoly = NULL;
}

#ifdef DEBUG
// Debug output for this edge
log4cpp::CategoryStream& operator<<(log4cpp::CategoryStream &os, 
				    const LMTEdge &edge) {
  os << *(edge.rev->dest) << " -> " << *(edge.dest);
  return os;
}
#endif // DEBUG
