// Declaration of the LMTEdge class, which represents a
// half-edge in the LMT-Skeleton.
#ifndef _LMTEDGE_H
#define _LMTEDGE_H

#ifdef DEBUG
#include <log4cpp/Category.hh>
#include <log4cpp/NDC.hh>
#endif // DEBUG

#include "LMTTypes.h"

class LMTPoint;

class LMTEdge {
 public:
  // The total number of groups so far.
  static int groupCounter;
  // The group this edge is in.
  int groupNumber;
  // Standard constructor: Creates a new half-edge with destination d.
  LMTEdge(LMTPoint *d)
    { dest = d; status = possible; rev = NULL; partnerFail = false;
    onStack = true; trianglesLeft = false; certificateChecked = false;
      visited = false; diamondRight = false; diamondLeft = false;
    leftPoly = NULL; rightPoly = NULL; groupNumber = 0;};
  // Advance the triangle to the left of the half-edge.
  bool advance();
  // Reset the triangle to the left of the half-edge.
  void reset();
  // Perform the dynamic programming step.
  void scan(int n);
  // Check whether this half-edge has a valid certificate.
  bool checkCertificate();
  // Check whether this half-edge is intersected by another edge.
  bool checkCrossingEdges();
  // Check whether this half-edge and edge intersect.
  bool intersects(LMTEdge *edge);
  // Restack all the edges whose certificates become invalid when 
  // this edge becomes impossible on edgeStack.
  void restackEdges(std::vector<LMTEdge *> &edgeStack);
  // Get the next edge in the cyclic ordering of edges around an endpoint,
  // performing lazy deletion.
  LMTEdge *getNext();
  // Initialize this edge for traversal.
  void traverse(const int counter);
public:
  // The destination of this half-edge.
  LMTPoint *dest;
  // The corresponding reverse edge.
  LMTEdge  *rev;
  // The current status of this edge.
  LMTStatus status;
  // The next edge in the cyclic ordering of the edges around an endpoint.
  LMTEdge   *next;
  // Is there a point in the left diamond of this edge?
  bool      diamondLeft;
  // Is there a point in the right diamond of this edge?
  bool      diamondRight;
  // Did the other half-edge fail the diamond test?
  bool      partnerFail;
  // The left edge of the certificate.
  LMTEdge   *i;
  // The right edge of the certificate.
  LMTEdge   *j;
  // The edge where the advance-process starts.
  LMTEdge   *dstart;
  // Is this edge currently on the edge stack?
  bool      onStack;
  // Has this edge already been visited during the face traversal?
  bool visited;
  // The right edge in the MWT-triangle to the left of this half-edge.
  LMTEdge   *rightPoly;
  // The left edge in the MWT-triangle to the left of this half-edge.
  LMTEdge   *leftPoly;
  // The optimal weight of the polygon to the left of this edge in the
  // LMT skelton. Used in the dynamic programming step.
  long double polyWeight;
  // The length of this edge.
  long double    length;
  // Index of destination vertex on the face during the dynamic programming
  // step.
  int destIndex;
 private:
  // Has checkCertificate already been called for this half-edge?
  bool certificateChecked;
  // Are there any triangles left to the left of this edge?
  bool trianglesLeft;
};

#ifdef DEBUG
// Debug output for this edge.
log4cpp::CategoryStream& operator<<(log4cpp::CategoryStream &os, 
				    const LMTEdge &edge);
#endif // DEBUG

#endif //_LMTEDGE_H
