// --------------------------------------------------------------------
// Ipelet for computing the LMT-Skeleton
// --------------------------------------------------------------------

#include <cstdlib>
#include <cstdio>
#include <cassert>

#include <ipelib.h>

#include "heuristic.h"
#include "LMTPoint.h"
#include "LMTEdge.h"

#ifdef __BORLANDC__
#pragma warn -8060
#endif

// --------------------------------------------------------------------

class LMTSkeletonIpelet : public Ipelet {
public:
  LMTSkeletonIpelet();
  virtual int IpelibVersion() const { return IPELIB_VERSION; }
  virtual int NumFunctions() const { return 2; }
  virtual const char *Label() const { return "LMT Skeleton"; }
  virtual const char *SubLabel(int function) const;
  virtual void Run(int function, IpePage *page, IpeletHelper *helper);
private:

private:
  std::vector<IpeVector> iSites;
};

LMTSkeletonIpelet::LMTSkeletonIpelet() {
}

const char * const sublabel[] = {
  "With DP",
  "Pure"
 };

const char *LMTSkeletonIpelet::SubLabel(int function) const
{
  return sublabel[function];
}

// --------------------------------------------------------------------

class CollectVisitor : public IpeVisitor {
public:
  CollectVisitor(std::vector<IpeVector> &sites);
  virtual void VisitGroup(const IpeGroup *obj);
  virtual void VisitPath(const IpePath *obj);
  virtual void VisitMark(const IpeMark *obj);
  virtual void VisitReference(const IpeReference *obj);
private:
  std::vector<IpeVector> &iSites;
  std::list<IpeMatrix> iStack;
};

CollectVisitor::CollectVisitor(std::vector<IpeVector> &sites)
  : iSites(sites)
{
  iStack.clear();
  iStack.push_back(IpeMatrix()); // id matrix
}

void CollectVisitor::VisitGroup(const IpeGroup *obj)
{
  iStack.push_back(iStack.back() * obj->Matrix());
  for (IpeGroup::const_iterator it = obj->begin(); it != obj->end(); ++it)
    (*it)->Accept(*this);
  iStack.pop_back();
}

void CollectVisitor::VisitPath(const IpePath *obj)
{
  IpeMatrix m = iStack.back() * obj->Matrix();
  for (int i = 0; i < obj->NumSubPaths(); ++i) {
    const IpeSegmentSubPath *sp = obj->SubPath(i)->AsSegs();
    if (sp) {
      iSites.push_back(m * sp->Segment(0).CP(0));
      for (int j = 0; j < sp->NumSegments(); ++j) {
	IpePathSegment seg = sp->Segment(j);
	iSites.push_back(m * seg.CP(seg.NumCP() - 1));
      }
    }
  }
}

void CollectVisitor::VisitMark(const IpeMark *obj)
{
  iSites.push_back(iStack.back() * obj->Matrix() * obj->Position());
}

void CollectVisitor::VisitReference(const IpeReference *obj)
{
  if (obj->Object()) {
    iStack.push_back(iStack.back() * obj->Matrix());
    obj->Object()->Accept(*this);
    iStack.pop_back();
  }
}

// --------------------------------------------------------------------


void LMTSkeletonIpelet::Run(int function, IpePage *page, 
			    IpeletHelper *helper) {

  // Collect the sites.
  iSites.clear();
  CollectVisitor vis(iSites);
  for (IpePage::const_iterator it = page->begin(); it != page->end(); ++it) {
    if (it->Select())
      vis(*it);
  }
  if (iSites.size() < 3) {
    helper->MessageBox("You need to select at least three sites.",
		       "Dismiss", 0, 0);
    return;
  }

 
  // Convert the sites to LMTPoints.
  std::vector<LMTPoint> points;
  for (uint i = 0; i < iSites.size(); i++) {
    LMTPoint p(iSites[i].iX, iSites[i].iY);
    points.push_back(p);
  }

  std::vector<LMTEdge *> iLMTSkeleton;
  std::vector<LMTEdge *> iCandidates;
  long double weight = LMTSkeleton(points, function,
			      iLMTSkeleton, iCandidates);

  if (!iLMTSkeleton.empty()) {
    // create edges
    IpeGroup *group = new IpeGroup;
    IpeAllAttributes attr; // all null attributes
    
    std::vector<IpeGroup *> groups;
    for (int k = 0; k <= LMTEdge::groupCounter; k++) {
      groups.push_back(new IpeGroup);
    }
    for (uint j = 0; j < iLMTSkeleton.size(); j++) {
      
      LMTEdge *currentEdge = iLMTSkeleton[j];
      IpeSegment seg(IpeVector(currentEdge->rev->dest->x, 
			       currentEdge->rev->dest->y), 
		     IpeVector(currentEdge->dest->x, 
			       currentEdge->dest->y));
      groups[currentEdge->groupNumber]->push_back(new IpePath(attr, seg));
    }
    for (uint k = 0; k < groups.size(); k++) {
      IpeGroup *currentGroup = groups[k];
      currentGroup->SetStroke(helper->Attributes().iStroke);
      currentGroup->SetDashStyle(helper->Attributes().iDashStyle);
      IpeAttribute b = IpeAttribute(IpeAttribute::ELineWidth, true, 1);
      currentGroup->SetLineWidth(b);
      group->push_back(currentGroup);
    }
    group->SetStroke(helper->Attributes().iStroke);
    group->SetDashStyle(helper->Attributes().iDashStyle);
    IpeAttribute b = IpeAttribute(IpeAttribute::ELineWidth, true, 1);
    group->SetLineWidth(b);
    page->push_back(IpePgObject(IpePgObject::ESecondary,
				helper->CurrentLayer(), group));
  }
  if (!iCandidates.empty()) {
    // create edges
    IpeGroup *group2 = new IpeGroup;
    IpeAllAttributes attr; // all null attributes
    IpeAttribute a = IpeAttribute(IpeAttribute::EDashStyle, true, 1);
    
    for (uint j = 0; j < iCandidates.size(); j++) {
      LMTEdge *currentEdge = iCandidates[j];
      IpeSegment seg(IpeVector(currentEdge->rev->dest->x, 
			       currentEdge->rev->dest->y), 
		     IpeVector(currentEdge->dest->x, 
			       currentEdge->dest->y));
      group2->push_back(new IpePath(attr, seg));
    }
    
    group2->SetStroke(helper->Attributes().iStroke);
    group2->SetDashStyle(a);
    group2->SetLineWidth(helper->Attributes().iLineWidth);
    page->push_back(IpePgObject(IpePgObject::ESecondary,
				helper->CurrentLayer(), group2));
  }
  iCandidates.clear();
  iLMTSkeleton.clear();
  iSites.clear();
  for (uint i = 0; i < points.size(); i++) {
    points[i].edgeList.clear();
  }
  points.clear();
  iLMTSkeleton.clear();
  char msg[200];
  sprintf(msg, "Weight is %f.", (double)weight);
  helper->Message(msg);
}

// --------------------------------------------------------------------

IPELET_DECLARE Ipelet *NewIpelet()
{
  return new LMTSkeletonIpelet;
}

// --------------------------------------------------------------------
