// A reimplementation of Ronald Beirouti's LMT-Heuristic,
// R. Beirouti and J. Snoeyink, 
// Implementations of the LMT Heuristic for Minimum Weight Triangulation, 
// Proceedings of the fourteenth annual symposium on Computational geometry,
// Minneapolis, Minnesota, United States,
// pp. 96 - 105.

#include <cstdlib>
#include <cstdio>
#include <cassert>
#include <cfloat>
#include <cmath>
#include <vector>
#include <list>
#include <algorithm>

#include "heuristic.h"

// Compare edges according to their length.
bool compareEdgesByLength(const LMTEdge *e1, const LMTEdge *e2) {
  return e1->length < e2->length;
} 

// Compare two edges during the dynamic programming step. The length
// of an ege is measured as the number of vertices between its endpoints
// on the polygon.
class TraversalComparator {
public:
  // Standard constructor, sets n.
  TraversalComparator(int _n) : n(_n) { }
  
  // Compare two edges according to the number of vertices between them.
  bool operator()(const LMTEdge *e1, const LMTEdge *e2) const {
    int diff1 = (e1->destIndex - e1->rev->destIndex + n) % n;
    int diff2 = (e2->destIndex - e2->rev->destIndex + n) % n;
    return diff1 < diff2;
  }
private:
  // The number of vertices of the polygon.
  int n;
};

// Perform BFS on the LMT-Skeleton in order to determine 
// whether the LMT-skeleton is connected.
bool isSkeletonConnected(std::vector<LMTPoint> &points) {
  std::list<LMTPoint *> queue;
  // start out with the first vertex
  queue.push_back(&(points[1]));
  points[1].visited = true;
  while (!queue.empty()) {
    // For each vertex, visit all the unvisited neighbors.
    LMTPoint *curPoint = queue.front();
    queue.pop_front();
    std::vector<LMTEdge> &edges = curPoint->edgeList;
    for (uint i = 0; i < edges.size(); i++) {
      if ((edges[i].status == certain) && !(edges[i].dest->visited)) {
	edges[i].dest->visited = true;
	queue.push_back(edges[i].dest);
      }
    } 
  }
  queue.clear();
  // If there remains an unvisited vertex, the skeleton is not connected.
  for (uint i = 0; i < points.size(); i++) {
    if (!points[i].visited) {
      return false;
    }
  }
  return true;
}

// Mark the edges that were determined by the dynamic programming step as
// certain.
void makeCertain(LMTEdge *edge) {
#ifdef DEBUG
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  logger.infoStream() << "  Making certain " << *edge 
		      << "(" << edge->polyWeight << ", ";
  //		      << *(edge->leftPoly) << ", " 
  //	      << *(edge->rightPoly) <<".";
#endif
  // If the edge is not marked as certain yet, make it certain and 
  // assign it to the appropriate group.
  if (edge->status != certain) {
    edge->status = edge->rev->status = certain;
    edge->visited = edge->rev->visited = true;
    edge->groupNumber = edge->rev->groupNumber = LMTEdge::groupCounter;
  }
  // If there is a left subpolygon, recurse there.
  if (edge->leftPoly != NULL && edge->leftPoly->status != certain) {
    makeCertain(edge->leftPoly);
  }
  // If there is a right subpolygon, recurse there.
  if (edge->rightPoly != NULL && edge->rightPoly->status != certain) {
    makeCertain(edge->rightPoly);
  }
}

// Traverse all the faces of the LMT-Skeleton and perform dynamic
// programming in order to determine the optimal triangulation.
// The result is the weight of the optimal triangulation.
long double traverseFaces(std::vector<LMTPoint> &points) {
  long double result = 0;
#ifdef DEBUG
  initLogging();
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  log4cpp::NDC::push("traverseFaces");
#endif
  // Iterate over all the certain half-edges.
  for (uint i = 0; i < points.size(); i++) {
    for (uint j = 0; j < points[i].edgeList.size(); j++) {
      if (points[i].edgeList[j].status == certain && 
	  !points[i].edgeList[j].visited) {
        // Start from this half-edge.
	LMTEdge *start = &(points[i].edgeList[j]);
	start->visited = true;
#ifdef DEBUG
	logger.infoStream() << "Visiting face starting with " << *start 
			    << "."; 
#endif
        // Counts the number of vertices on the current face.
	int counter = 0;
        // Visit the start edge, and set the index of the destination
        // vertex.
	start->traverse(counter);
	start->destIndex = ++counter;
        // Stores all the edges to be considered during the dynamic programming
        // setp.
	std::vector<LMTEdge *> candidateEdges;
	candidateEdges.push_back(start->rev);
        // Finde the next edge along the face, and keep track of all the
        // possible edges encountered..
	LMTEdge *current = start->rev->next;
	current->traverse(counter);
	while (current->status != certain) {
	  if (current->status == possible) {
	    candidateEdges.push_back(current);
	  }
	  current = current->next;
	  current->traverse(counter);
	}
        // Continue until we have traversed the face.
	while (current != start) {
          // Initialize the current skeleton edge.
	  candidateEdges.push_back(current->rev);
	  current->destIndex = ++counter;
	  current->visited = true;
#ifdef DEBUG
	  logger.infoStream() << "  visiting  " << *current << "."; 
#endif
          // Find the next skeleton edge, storing the possible edges
          // encountered.
	  current = current->rev->next;
	  current->traverse(counter);
	  while (current->status != certain) {
	    if (current->status == possible) {
	      candidateEdges.push_back(current);
	    }
	    current = current->next;
	    current->traverse(counter);
	  }
	}
        // Sort the candidate edges according to the number of vertices
        // between their endpoints.
	std::sort(candidateEdges.begin(), 
		  candidateEdges.end(),
		  TraversalComparator(counter));
        // If there are more than three edges on the face, we need to
        // do dynamic programming.
	if (counter > 3) {
#ifdef DEBUG
	  logger.infoStream() << "  Candidates:";
	  for (uint k = 0; k < candidateEdges.size(); k++) {
	    logger.infoStream() << "    " << *(candidateEdges[k])
				<< " " << candidateEdges[k]->rev->destIndex
				<< " -> " << candidateEdges[k]->destIndex;
	  }
#endif
          // Scan all the edges in order of their length.
	  for (uint k = 0; k < candidateEdges.size(); k++) {
	    LMTEdge *curEdge = candidateEdges[k];
	    int dist = 
	      (curEdge->destIndex - curEdge->rev->destIndex + counter) % counter;
	    if (dist <= 2) {
	      curEdge->polyWeight = 0;
	    } else {
	      curEdge->scan(counter);
	    }
	  }
#ifdef DEBUG
	  logger.infoStream() << "  Scanning complete:";
	  for (uint k = 0; k < candidateEdges.size(); k++) {
	    logger.infoStream() << "    " << *(candidateEdges[k])
				<< " " << candidateEdges[k]->polyWeight << " " 
				<< ((candidateEdges[k]->leftPoly != NULL) ?
				    *(candidateEdges[k]->leftPoly) 
				     : 
				    *(candidateEdges[k])) << " " 
				<< (candidateEdges[k]->rightPoly != NULL ?
				    *(candidateEdges[k]->rightPoly)
				    : 
				    *(candidateEdges[k]));
	  }
#endif
          // Add the triangulation determined by the dynamic programming
          // step to the triangulation. Every face gets its own group
          // number.
	  if (start->rev->polyWeight != DBL_MAX) {
	    LMTEdge::groupCounter++;
	    makeCertain(start->rev);
	    result += start->rev->polyWeight;
	  }
	}
	candidateEdges.clear();
      }
    }
  }
  return result;
}

// Perfrom the LMT-Skeleton heuristic on points, and report the
// skeleton and candiate edges back. If func is 0, dynamic programming
// will be performed if the skeleton is connected. The result is the
// total length of the resulting triangulation.
long double LMTSkeleton(std::vector<LMTPoint> &points, int func,
		 std::vector<LMTEdge *> &skeleton,
		 std::vector<LMTEdge *> &candidates) {
  long double result;
#ifdef DEBUG
  initLogging();
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  log4cpp::NDC::push("LMTSkeleton");
#endif
  // Initialize the data structure
  initDataStructure(points);
#ifdef DEBUG
  dumpDataStructure(points);
#endif
  // Initialize and sort the edge stack
  std::vector<LMTEdge *> edgeStack;
  for (uint i = 0; i < points.size(); i++) {
    for (uint j = 0; j < points[i].edgeList.size(); j++) {
      LMTEdge *currentEdge = &(points[i].edgeList[j]);
      if (*(currentEdge->rev->dest) < *(currentEdge->dest)) {
	edgeStack.push_back(currentEdge);
      }
    }
  }
  std::sort(edgeStack.begin(), edgeStack.end(), compareEdgesByLength);
#ifdef DEBUG
  logger.infoStream() << "There are " << edgeStack.size() 
		      << " edges on the stack.";
#endif
  // Perfom the heuristic
  // Check all the edges on the stack.
  while (!edgeStack.empty()) {
    // Get the next edge.
    LMTEdge *currentEdge = edgeStack.back();
    edgeStack.pop_back();
    currentEdge->onStack = currentEdge->rev->onStack = false;
    if (!currentEdge->checkCertificate()) {
      if (!currentEdge->checkCrossingEdges() && 
	  !currentEdge->rev->checkCrossingEdges()) {
	// If the current edge has no certificate, but is not crossed by 
        // other edges, it must be certain.
	currentEdge->status = currentEdge->rev->status = certain;
	currentEdge->groupNumber = currentEdge->rev->groupNumber = LMTEdge::groupCounter;
      } else {
        // If the current edge has no certificate and is crossed by
        // another edge, it cannot be part of the LMTSkeleton.
	currentEdge->restackEdges(edgeStack);
	currentEdge->status = currentEdge->rev->status = impossible;
      }
    }
  }
  edgeStack.clear();
#ifdef DEBUG
  logger.infoStream() << "Postprocessing step.";
#endif
  // Mark all the edges that have a certificate but no
  // crossing edges as certain.
  for (uint i = 0; i < points.size(); i++) {
    for (uint j = 0; j < points[i].edgeList.size(); j++) {
      LMTEdge *currentEdge = &(points[i].edgeList[j]);
#ifdef DEBUG
      logger.infoStream() << "Checking " << (*currentEdge) << ".";
#endif
      if (currentEdge->status == possible && 
	  !currentEdge->checkCrossingEdges() && 
	  !currentEdge->rev->checkCrossingEdges()) {
#ifdef DEBUG
      logger.infoStream() << "Edge " << (*currentEdge) << " is certain";
#endif
	  currentEdge->status = currentEdge->rev->status = certain;
	  currentEdge->groupNumber = currentEdge->rev->groupNumber = LMTEdge::groupCounter;
      }
    }
  }
  // Check if the LMTSkeleton is connected.
  bool connected = isSkeletonConnected(points);
  if (func != 0) {
    connected = false;
  }
  // If it is, perform dynamic programming to determine the minimum weight
  // triangulation.
  if (connected) {
#ifdef DEBUG
  logger.infoStream() << "The LMT-Skeleton is connected.";
#endif
    traverseFaces(points);
  } else {
#ifdef DEBUG
  logger.infoStream() << "The LMT-Skeleton is not connected.";
#endif
  }
  // Add all the certain edges and the candidate edges to the skeleton.
  result = 0;
  //  std::vector<LMTEdge *> result;
  for (uint i = 0; i < points.size(); i++) {
    for (uint j = 0; j < points[i].edgeList.size(); j++) {
      LMTEdge *currentEdge = &(points[i].edgeList[j]);
      if (*(currentEdge->rev->dest) < *(currentEdge->dest)) {
	if (currentEdge->status == certain) {
#ifdef DEBUG
	  logger.infoStream() << "Skeleton:" << (*currentEdge) << ".";
#endif
	  skeleton.push_back(currentEdge);
	  result += currentEdge->length;
	} else if (!connected & currentEdge->status == possible) {
#ifdef DEBUG
	  logger.infoStream() << "Candidate:" << (*currentEdge) << ".";
#endif
	  candidates.push_back(currentEdge);
	}
      } 
    }
  }
#ifdef DEBUG
  log4cpp::NDC::pop();
#endif
  return result;
}
