// Initialize the LMT-Skeleton datastructure.

#include <cassert>
#include <cmath>
#include <cstdlib>
#include <vector>
#include "init.h"
#include "util.h"
#include <algorithm>

// The base angle for the diamond test.
#define DIAMOND_ANGLE cos(M_PI/4.6)

// Compares to points in polar order with repsect to a given origin.
class compPolar {
public:
  // Standard constructor, provides the origin and a flag which indicates
  // whether to check for collinearity or not.
  compPolar(LMTPoint _origin, bool _checkColl) { 
    origin = _origin; 
    checkColl = _checkColl;
  }
  
  // Compare two points in polar order wrt to origin. If the collinearity flag
  // is set and the two points are collinear, the point which is closer to 
  // the origin will be considered smaller.
  int operator() (const LMTPoint &p1, const LMTPoint &p2) const {
    int quad1 = quadrant(p1, origin);
    int quad2 = quadrant(p2, origin);
    if (quad1 != quad2) {
      return quad1 < quad2;
    } else {
      return isRightOf(origin, p2, p1)
	       ||
	(checkColl && (isCollinear(origin, p2, p1) && lengthOf(origin, p1) < 
	 lengthOf(origin, p2)));
    }
  }

  // Compare the endpoints of two edges in polar order wrt to origin.
  int operator() (const LMTEdge &e1, const LMTEdge &e2) const {
    return (*this)(*(e1.dest), *(e2.dest));
  }

  // Get the origin for this comparator.
  LMTPoint *getOrigin() {
    return &(this->origin);
  }
private:
  // The origin for the comparator.
  LMTPoint origin;
  // Do we need to check for collinearity?
  bool checkColl;
};

// Checks whether two edges are collinear. 
// The start-point of the edges needs to be provided explicitly,
// since the rev-pointers have not been set up yet.
class compCollinear {
public:
  // Standard constructor, provides the origin.
  compCollinear(LMTPoint _start) { start = _start; }

  // Check whether e1 and e2 are collinear.
  bool operator() (const LMTEdge &e1, const LMTEdge &e2) const {
    LMTPoint p1 = *(e1.dest);
    LMTPoint p2 = *(e2.dest);
    int quad1 = quadrant(p1, start);
    int quad2 = quadrant(p2, start);
    if (quad1 != quad2) {
      return false;
    } else {
      return isCollinear(start, p2, p1);
    }
  }

private:
  // The start-point of the edges.
  LMTPoint start;
};


// Apply the diamond test. That is, check whether the endpoint of 
// piovot lies in the diamond around e1. If so, set the appropriate
// flag in e1. The start-point of e1 and pivot needs to be provided
// explicitly, since the rev-pointers have not been set up yet.
void applyDiamondTest(LMTEdge &e1, const LMTEdge &pivot, 
		      const LMTPoint *start) {
#ifdef DEBUG
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  logger.infoStream() << "Diamond test for " << *e1.dest << " (Testing"
		      << *pivot.dest << ").";
#endif
  coordT x1 = e1.dest->x - start->x;
  coordT y1 = e1.dest->y - start->y;
  coordT x2 = pivot.dest->x - start->x;
  coordT y2 = pivot.dest->y - start->y;
  coordT x3 = pivot.dest->x - e1.dest->x;
  coordT y3 = pivot.dest->y - e1.dest->y;
  coordT l3 = sqrt(x3*x3+y3*y3);
  coordT sprod1 = (x1*x2 + y1*y2) / (e1.length * pivot.length);
  coordT sprod2 = (-x1*x3 - y1*y3) / (e1.length * l3);
  if (sprod1 >= cos(DIAMOND_ANGLE) && sprod2 >= cos(DIAMOND_ANGLE)) {
    if (isLeftOf(*start, *(e1.dest), *(pivot.dest))) {
      e1.diamondLeft = true;
    } else if (isRightOf(*start, *(e1.dest), *(pivot.dest))) {
      e1.diamondRight = true;
    } else {
      //      assert(false);
    }
#ifdef DEBUG
    logger.infoStream() << "Test positive (" 
			<< (e1.diamondLeft ? "left" : "right") << ").";
#endif
  } 
}

// Swap two element in edges. This function is used in the quicksort
// algorithm.
inline void swap(std::vector<LMTEdge> &edges, int p1, int p2) {
  LMTEdge temp = edges[p1];
  edges[p1] = edges[p2];
  edges[p2] = temp;
}

// Perfrom the quicksort algorithm to sort the elements in [l,r)
// in the vector edges. Comparisons are performed by lessThan.
// During partitioning, we apply the diamond test.
void qs(std::vector<LMTEdge> &edges, int l, int r, compPolar &lessThan) {
  while (l < r) {
    // Choose the shortest edge as the pivot element and put it into the
    // right position.
    int pivot = l;
    for (int i = l + 1; i < r; i++) {
      if (edges[i].length < edges[pivot].length) {
	pivot = i;
      }
    }
    swap(edges, pivot, l);
    // Perform the partition step, apply the diamond test. 
    int lastSmaller = l;
    int pos = l+1;
    while (pos < r) {
      applyDiamondTest(edges[pos], edges[l], lessThan.getOrigin());
      if (lessThan(edges[pos], edges[l])) {
	lastSmaller++;
	swap(edges, lastSmaller, pos);
      }
      pos++;
    }
    swap(edges, lastSmaller, l);
    // Recurse.
    if (lastSmaller - l < r - lastSmaller) {
      qs(edges, l, lastSmaller, lessThan);
      l = lastSmaller + 1;
    } else {
      qs(edges, lastSmaller + 1, r, lessThan);
      r = lastSmaller;
    }
  }
}

// Perform quicksort.
void quicksort(std::vector<LMTEdge> &edges, compPolar &lessThan) {
  qs(edges, 0, edges.size(), lessThan);
}

// Check whether edge fails the diamond test. This is the case if there
// are points in left and the right half of edge's diamond, or if the
// twin half-edge of this edge failed the diamond test.
bool failDiamond(const LMTEdge &edge) {
  bool result = (edge.diamondLeft && edge.diamondRight) || edge.partnerFail;
#ifdef DEBUG
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  logger.infoStream() << "  Edge " << *(edge.dest) << 
    (result ? " failed." : " passed.");
#endif
  return result;
} 

// Find the twin half-edges for all the edges going out of a and
// update the information about the diamond test. The rev pointers 
// are not actually set yet, because the edges that fail the diamond
// test need to be delted first.
void checkTwinDiamond(LMTPoint &a) {
#ifdef DEBUG
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  log4cpp::NDC::push("findPartners");
  logger.infoStream() << "Finding partners for " << a << ".";
#endif
  // Check all edges.
  for (uint j = 0; j < a.edgeList.size(); j++) {
    // Find the twin of this half edge in the edgelist of the current 
    // edges end-point.
    std::vector<LMTEdge> &reverseEdges = a.edgeList[j].dest->edgeList;
    if (reverseEdges.size() == 0) {
      continue;
    }
    LMTEdge dummy(&a);
    std::vector<LMTEdge>::iterator 
      iter = std::lower_bound(reverseEdges.begin(),
			      reverseEdges.end(),
			      dummy, compPolar(*(a.edgeList[j].dest), false));
    // Set the information about the diamond test.
    if (iter == reverseEdges.end() || iter->dest != &a) {
      a.edgeList[j].partnerFail = true;
    } else if (iter->diamondLeft && iter->diamondRight) {
	a.edgeList[j].partnerFail = true;
    } else if (a.edgeList[j].diamondLeft && a.edgeList[j].diamondRight) {
      iter->partnerFail = true;
    }
  }
#ifdef DEBUG
  logger.info("Done.");
  log4cpp::NDC::pop();
#endif
}

// Initialize the data-structure for sites.
void initDataStructure(std::vector<LMTPoint> &sites) {
#ifdef DEBUG
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  log4cpp::NDC::push("initDS");
  logger.info("Initializing data structure.");
  logger.infoStream() << "There are " << sites.size() << " points.";
#endif
  // Remove multiple points.
  std::sort(sites.begin(), sites.end());
  std::vector<LMTPoint>::iterator iter = 
    std::unique(sites.begin(), sites.end());
  sites.erase(iter, sites.end());
#ifdef DEBUG
  logger.infoStream() << "There are " << sites.size() << " points remaining.";
  logger.info("Initializing edges.");
#endif
  // Create the edges for each site,
  for (uint i = 0; i < sites.size(); i++) {
    LMTPoint &a = sites[i];
#ifdef DEBUG
    logger.infoStream() << "Initializing edges for node " << i << 
      " " << a << "."; 
#endif
    // Create all edges leaving a.
    for(uint j = 0; j < sites.size(); j++) {
      LMTPoint &b = sites[j];
      if (i != j) {
	LMTEdge edge(&b);
	edge.length = lengthOf(a, b);
	a.edgeList.push_back(edge);
      }
    }
#ifdef DEBUG
    logger.infoStream() << "There are " << a.edgeList.size() << " edges."; 
    logger.info("Sorting edges.");
#endif
    // Sort the edges radially around a and perform the diamond test.
    compPolar comparator = compPolar(a, true);
    quicksort(a.edgeList, comparator);
#ifdef DEBUG
    logger.info("Sorted.");
#endif
    // If there are collinear edges, only keep the shortest edge.
    std::vector<LMTEdge>::iterator iter 
      =  std::unique(a.edgeList.begin(), a.edgeList.end(), compCollinear(a));
    a.edgeList.erase(iter, a.edgeList.end());
    // Update the information about the diamond test.
    checkTwinDiamond(a);
#ifdef DEBUG
    logger.infoStream() << "There are " << a.edgeList.size()
			<< " edges left."; 
#endif
  }
  // Delete all the edges that failed the diamond test.
  for (uint i = 0; i < sites.size(); i++) {
    LMTPoint &a = sites[i];
#ifdef DEBUG
    logger.infoStream() << "Reconsidering node " << i << 
      " " << a << "."; 
    logger.infoStream() << "Deleting diamond edges.";
#endif
    std::vector<LMTEdge>::iterator iter = 
      std::remove_if(a.edgeList.begin(), a.edgeList.end(), failDiamond);
    a.edgeList.erase(iter, a.edgeList.end());
#ifdef DEBUG
    logger.infoStream() << "There are " << a.edgeList.size()
			<< " edges left."; 
#endif
  }
  // Set up the next and rev pointers.
  for (uint i = 0; i < sites.size(); i++) {
    LMTPoint &a = sites[i];
#ifdef DEBUG
    logger.infoStream() << "Rereconsidering node " << i << 
      " " << a << "."; 
    logger.info("Setting rev pointers.");
#endif
    for (uint j = 0; j < a.edgeList.size(); j++) {
#ifdef DEBUG
      logger.infoStream() << "Considering edge " << j << ".";
      logger.infoStream() << "Endpoint is " << *(a.edgeList[j].dest) << ".";
#endif
      if (a.edgeList[j].rev != NULL) continue;
      // Look for the twin of this edge in the destion vertex' edge list,
      // end set the reverse pointers.
      std::vector<LMTEdge> &reverseEdges = a.edgeList[j].dest->edgeList;
      LMTEdge dummy(&a);
      std::vector<LMTEdge>::iterator 
	iter = std::lower_bound(reverseEdges.begin(),
				reverseEdges.end(),
				dummy, compPolar(*(a.edgeList[j].dest), false));
      if (iter == reverseEdges.end() || iter->dest != &a) {
#ifdef DEBUG
	logger.infoStream() << "  No partner found.";
#endif
	assert(false);
      } else {
#ifdef DEBUG
	logger.infoStream() << "  Partner has endpoint "  << *(iter->dest) 
			    << ".";
#endif
	iter->rev = &(a.edgeList[j]);
	a.edgeList[j].rev = &(*iter);
      }
    }
#ifdef DEBUG
    logger.info("Setting next pointers.");
#endif
    // Set the next pointers for all the edges leaving a.
    for (uint j = 0; j < a.edgeList.size(); j++) {
      a.edgeList[j].next = &(a.edgeList[(j + 1) % a.edgeList.size()]);
    }
  }
#ifdef DEBUG
  logger.info("Initializing dstart-pointers.");
#endif
  // Initialize the dstart-pointers for each node.
  for (uint i = 0;  i < sites.size(); i++) {
#ifdef DEBUG
    logger.infoStream() << "Setting dstart-pointers for node " << i <<".";
#endif
    LMTPoint &a = sites[i];
    if (a.edgeList.size() == 0) {
      continue;
    }
    // The first edge.
    LMTEdge *startEdge = &(a.edgeList[0]);
    // The current edge for whose twin the dstart pointer is set.
    LMTEdge *currentEdge = startEdge;
    // The edge which will become the dstart edge.
    LMTEdge *opposedEdge = startEdge;
    do {
#ifdef DEBUG
      logger.infoStream() << "Current edge has endpoint " << 
	*(currentEdge->dest) << "."; 
#endif
      // Advance the opposed edge until its endpoint it to the right of the
      // current edge.
      while (opposedEdge->next->dest->isLeftOf(currentEdge)) {
	opposedEdge = opposedEdge->next;
      }
#ifdef DEBUG
      logger.infoStream() << "Opposed edge has endpoint " << 
	*(opposedEdge->dest) << ".";
#endif
      // Set the dstart pointer for the current edge's twin and advance
      // current edge.
      currentEdge->rev->dstart = opposedEdge;
      currentEdge = currentEdge->next;
    } while(currentEdge != startEdge);
  }
#ifdef DEBUG
  logger.info("Done.");
  log4cpp::NDC::pop();
#endif
}

#ifdef DEBUG
// Output debug information.
void dumpDataStructure(std::vector<LMTPoint> &sites) {
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  log4cpp::NDC::push("dumpDS");
  logger.infoStream() << "There are " << sites.size() << " Points.";
  for (uint i = 0; i < sites.size(); i++) {
    logger.infoStream() << "Vertex " << i << " " << sites[i] << ".";
    logger.infoStream() << "There are " << sites[i].edgeList.size()
			<< " Edges.";
    for (uint j = 0; j < sites[i].edgeList.size(); j++) {
      logger.infoStream() << "Edge " << j << ": " <<
	sites[i].edgeList[j] << ".";
      logger.infoStream() << "  dstart: " <<
	*(sites[i].edgeList[j].dstart) << ".";
    }
  }
  log4cpp::NDC::pop();
}

log4cpp::FileAppender appender("LMT-Heuristik", "c:\\lmt.log");

// Initialize the logger.
void initLogging() {
  log4cpp::Category &logger = log4cpp::Category::getRoot();
  appender.setLayout(new log4cpp::BasicLayout);
  logger.addAppender(&appender);
}

#endif // DEBUG
